@@ -1,5 +1,5 @@
 package Devel::Camelcadedb;
-our $VERSION = "1.6.1.6";
+our $VERSION = "2017.1.1";
 
 # http://perldoc.perl.org/DB.html
 # http://perldoc.perl.org/perldebug.html
@@ -26,6 +26,21 @@ use constant {
     STEP_OVER     => 2,
 };
 
+use constant {
+    # see PERLDBf_* constants in perl.h
+    DEBUG_ALL             => 0x7ff,
+    DEBUG_SINGLE_STEP_ON  => 0x20,
+    DEBUG_USE_SUB_ADDRESS => 0x40,
+    DEBUG_REPORT_GOTO     => 0x80,
+};
+use constant {
+    # debugger enabled
+    DEBUG_DEFAULT_FLAGS => DEBUG_ALL & ~(DEBUG_USE_SUB_ADDRESS|DEBUG_REPORT_GOTO), # 0x73f
+    # instrument code, but don't call DB::DB (see sub disable for DB::sub)
+    DEBUG_PREPARE_FLAGS => DEBUG_ALL & ~(DEBUG_USE_SUB_ADDRESS|DEBUG_REPORT_GOTO|DEBUG_SINGLE_STEP_ON), # 0x73c
+};
+
+
 # Each array @{"::_<$filename"} holds the lines of $filename for a file compiled by Perl. The same is also true for evaled
 # strings that contain subroutines, or which are currently being executed. The $filename for evaled strings looks like
 # (eval 34) .
@@ -992,7 +1007,9 @@ sub _get_real_path_by_normalized_perl_file_id
     if (!exists $_perl_file_id_to_path_map{$perl_file_id})
     {
         no strict 'refs';
-        my $real_path = _calc_real_path( ${*{"::_<$perl_file_id"}}, $perl_file_id );
+        my $path = ${*{"::_<$perl_file_id"}};
+        return '' unless defined $path; # some subs created via XS
+        my $real_path = _calc_real_path( $path, $perl_file_id );
 
         $_perl_file_id_to_path_map{$perl_file_id} = $real_path;
         $_paths_to_perl_file_id_map{$real_path} = $perl_file_id;
@@ -1178,6 +1195,27 @@ sub _set_up_debugger
     }
 }
 
+sub _set_up_after_connect
+{
+    my ($allow_fail) = @_;
+
+    $_debug_socket->autoflush( 1 );
+    $_debug_socket_select = IO::Select->new();
+    $_debug_socket_select->add( $_debug_socket );
+
+    _send_data_to_debugger( +{
+        event   => 'READY',
+        version => $VERSION,
+    } );
+    _report "Waiting for set up data..." if $_dev_mode;
+    my $set_up_data = <$_debug_socket>;
+    return if !defined $set_up_data && $allow_fail;
+    die "Connection closed" unless defined $set_up_data;
+
+    $ready_to_go = 1;
+    $DB::single = _set_up_debugger( $set_up_data );
+}
+
 sub _process_new_breakpoints
 {
     my ($json_data) = @_;
@@ -1321,7 +1359,7 @@ sub _switch_context
 # feature is disabled when executing inside DB::DB() , including functions called from it unless $^D & (1<<30) is true.
 sub step_handler
 {
-    return if $_internal_process;
+    return if $_internal_process || !$ready_to_go;
     $_internal_process = 1;
 
     # Save eval failure, command failure, extended OS error, output field
@@ -1695,41 +1733,108 @@ EOM
     exit;
 }
 
-# http://perldoc.perl.org/perlipc.html#Sockets%3a-Client%2fServer-Communication
-if ($ENV{PERL5_DEBUG_ROLE} eq 'server')
+my $_connect_at_start = exists $ENV{PERL5_DEBUG_AUTOSTART} ? $ENV{PERL5_DEBUG_AUTOSTART} : 1;
+
+sub is_connected
 {
-    printf STDERR "Listening for the IDE connection at %s:%s...\n", $ENV{PERL5_DEBUG_HOST}, $ENV{PERL5_DEBUG_PORT};
-    my $_server_socket = IO::Socket::INET->new(
-        Listen    => 1,
-        LocalAddr => $ENV{PERL5_DEBUG_HOST},
-        LocalPort => $ENV{PERL5_DEBUG_PORT},
-        ReuseAddr => 1,
-        Proto     => 'tcp',
-    ) || die "Error binding to $ENV{PERL5_DEBUG_HOST}:$ENV{PERL5_DEBUG_PORT}";
-    $_debug_packed_address = accept( $_debug_socket, $_server_socket );
+    return !!$_debug_socket;
 }
-else
+
+sub _connect
 {
-    foreach my $attempt (1 .. 10)
+    my ($attempts, $allow_fail) = @_;
+    # http://perldoc.perl.org/perlipc.html#Sockets%3a-Client%2fServer-Communication
+    if ($ENV{PERL5_DEBUG_ROLE} eq 'server')
     {
-        printf STDERR "($attempt)Connecting to the IDE from process %s at %s:%s...\n", $$, $ENV{PERL5_DEBUG_HOST},
-            $ENV{PERL5_DEBUG_PORT};
-        $_debug_socket = IO::Socket::INET->new(
-            PeerAddr  => $ENV{PERL5_DEBUG_HOST},
-            PeerPort  => $ENV{PERL5_DEBUG_PORT},
+        printf STDERR "Listening for the IDE connection at %s:%s...\n", $ENV{PERL5_DEBUG_HOST}, $ENV{PERL5_DEBUG_PORT};
+        my $_server_socket = IO::Socket::INET->new(
+            Listen    => 1,
+            LocalAddr => $ENV{PERL5_DEBUG_HOST},
+            LocalPort => $ENV{PERL5_DEBUG_PORT},
             ReuseAddr => 1,
             Proto     => 'tcp',
-        );
-        last if $_debug_socket;
-        sleep( 1 ); # this is kinda hacky
+        ) || die "Error binding to $ENV{PERL5_DEBUG_HOST}:$ENV{PERL5_DEBUG_PORT}";
+        $_debug_packed_address = accept( $_debug_socket, $_server_socket );
     }
-    die "Error connecting to $ENV{PERL5_DEBUG_HOST}:$ENV{PERL5_DEBUG_PORT}" unless $_debug_socket;
+    else
+    {
+        foreach my $attempt (1 .. $attempts)
+        {
+            printf STDERR "($attempt)Connecting to the IDE from process %s at %s:%s...\n", $$, $ENV{PERL5_DEBUG_HOST},
+                $ENV{PERL5_DEBUG_PORT};
+            $_debug_socket = IO::Socket::INET->new(
+                PeerAddr  => $ENV{PERL5_DEBUG_HOST},
+                PeerPort  => $ENV{PERL5_DEBUG_PORT},
+                ReuseAddr => 1,
+                Proto     => 'tcp',
+            );
+            last if $_debug_socket || $attempt == $attempts;
+            sleep( 1 ); # this is kinda hacky
+        }
+        die "Error connecting to $ENV{PERL5_DEBUG_HOST}:$ENV{PERL5_DEBUG_PORT}" if !$_debug_socket && !$allow_fail;
+    }
+    _set_up_after_connect( $allow_fail ) if $_debug_socket;
 }
-$_debug_socket->autoflush( 1 );
-$_debug_socket_select = IO::Select->new();
-$_debug_socket_select->add( $_debug_socket );
 
-print STDERR "Connected.\n";
+sub connect
+{
+    _connect( 1, 1 );
+}
+
+sub disconnect
+{
+    return unless is_connected();
+    $_debug_socket->close();
+    undef $_debug_socket_select;
+    undef $_debug_socket;
+    $ready_to_go = 0;
+}
+
+sub connect_or_reconnect
+{
+    disconnect() if is_connected();
+    _connect( 1, 1 );
+}
+
+# we want disable() to completely bypass the debugger (except for the parts
+# which are required for bookkeeping, like DB::postponed)
+#
+# setting %^P can disable DB::DB, but the only way to disable DB::sub is to
+# make sure *DB::sub{CODE} is undef, while keeping %DB::sub and $DB::sub
+# intact; the only way to do that is to save the glob slots we want to
+# preserve, undef the glob and then restore the slots
+#
+# DB::lsub and DB::goto are easier, because we don't need to preserve the
+# corresponding scalar/array/hash variables
+my (%_orig_db_sub, %_disabled_db_sub, $_orig_db_lsub);
+
+BEGIN
+{
+    %_orig_db_sub = %_disabled_db_sub = map {
+        ($_ => *DB::sub{$_}) x !!*DB::sub{$_}
+    } qw(SCALAR ARRAY HASH);
+    $_orig_db_sub{CODE} = \&sub_handler;
+    $_orig_db_lsub = undef; # \&lsub_handler
+}
+
+sub enable
+{
+    $^P = DEBUG_DEFAULT_FLAGS;
+    undef *DB::sub;
+    *DB::sub = $_orig_db_sub{$_} for keys %_orig_db_sub;
+    #*DB::lsub = $_orig_db_lsub;
+    #*DB::goto = \&goto_handler;
+}
+
+sub disable
+{
+    $DB::single = 0;
+    $^P = DEBUG_PREPARE_FLAGS;
+    undef *DB::sub;
+    undef *DB::lsub;
+    undef *DB::goto;
+    *DB::sub = $_disabled_db_sub{$_} for keys %_disabled_db_sub;
+}
 
 push @$_stack_frames, {
         subname      => 'SCRIPT',
@@ -1757,25 +1862,15 @@ foreach my $main_key (keys %::)
     }
 }
 
-_send_data_to_debugger( +{
-        event   => 'READY',
-        version => $VERSION,
-    } );
-_report "Waiting for set up data..." if $_dev_mode;
-my $set_up_data = <$_debug_socket>;
-die "Connection closed" unless defined $set_up_data;
-
 *DB::DB = \&step_handler;
-*DB::sub = \&sub_handler;
-#*DB::lsub = \&lsub_handler;
 *DB::postponed = \&load_handler;
-#*DB::goto = \&goto_handler;
 
-my $initial_state = _set_up_debugger( $set_up_data );
+if ($_connect_at_start)
+{
+    _connect( 10, 0 );
+    enable();
+}
 
 $_internal_process = 0;
-$ready_to_go = 1;
-
-$DB::single = $initial_state;
 
 1; # End of Devel::Camelcadedb
